import 'dart:async';

import 'package:firebase_crashlytics/firebase_crashlytics.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:logging/logging.dart';
import 'package:pub_semver/pub_semver.dart';
import 'package:sunny_lifecycle/sunny_lifecycle.dart';
import 'package:sunny_lifecycle/user_preferences_service.dart';
import 'package:sunny_sdk_core/model_exports.dart';
import 'package:sunny_service_stubs/sunny_service_stubs.dart';

const analyticsIdKey = UserPrefKey("analyticsId");

AnalyticsService? _analytics;
AnalyticsService get analytics => _analytics ??= AnalyticsService._();

/// Any new provider can adapt this.
abstract class AnalyticsProvider {
  static Map<String, Object> sanitizeParams(Map<String, dynamic>? input) {
    if (input == null) return {};
    return {
      for (var entry in input.entries)
        if (entry.value != null) entry.key: sanitize(entry.value),
    };
  }

  static Object sanitize(dynamic input) {
    if (input is String || input is num || input is bool || input is DateTime) {
      return input as Object;
    }
    if (input is Iterable) {
      return [
        for (var item in input) sanitize(item),
      ];
    }
    if (input is Map<String, dynamic>) {
      return sanitizeParams(input);
    }
    _log.info("Don't know how to handle ${input.runtimeType} so we're toString()-ing it");
    return "$input";
  }

  void logPageView({String? route, String? label, String? resolvedPath, Map<String, dynamic>? params});

  void share({String? shareContentType, String? shareContentId, String? shareMethod, String? shareTarget});

  void identify({String? userId, String? username});

  void track(dynamic eventName, Map<String, dynamic> properties);

  void setUserProperties(Map<String, dynamic> properties);

  set email(String email);
  set name(String? name);

  void encounteredError(Object error, [StackTrace? stack, dynamic context]);

  void reset(String? anonymousId);
}

final StreamController<String> userIds = StreamController.broadcast();

List<AnalyticsProvider>? _defaultAnalyticsProviders;

set defaultAnalyticsProviders(List<AnalyticsProvider> providers) {
  if (_analytics != null) {
    _analytics!.providers.clear();
    _analytics!.providers.addAll(providers);
  } else {
    _defaultAnalyticsProviders = providers;
  }
}

List<AnalyticsProvider> get defaultAnalyticsProviders {
  return _defaultAnalyticsProviders ??
      <AnalyticsProvider>[
//  if (!kIsWeb) SunnyMixpanel(),
//   if (!kIsWeb) SunnyFirebase(),
        if (!kIsWeb) SunnyCrashlytics(),
      ];
}

dynamic analyticsValue(dynamic value, [int depth = 0]) {
  if (value == null) return null;

  if (value is num) return value;
  if (value is String) return value;
  return "$value";
}

final _log = Logger("routes");

class AnalyticsNavigatorObserver extends NavigatorObserver with LoggingMixin {
  final String root;

  AnalyticsNavigatorObserver([this.root = '/']);

  @override
  void didPush(Route<dynamic> route, Route<dynamic>? previousRoute) {
    // final name = route?.settings?.name;

    if (route.settings is PathRouteSettings) {
      final settings = route.settings as PathRouteSettings;
      _log.info("#####################################");
      if (settings.label != null) _log.info("Page: ${settings.label}");
      if (settings.route != null) _log.info("Route: ${settings.route}");
      final args = settings.routeParams?.toMap() ?? {};
      if (settings.resolvedPath != null) _log.info("Path: ${settings.resolvedPath}");
      if (args.isNotEmpty) _log.info("Args: ${args.keys.toList()}");
      _log.info("#####################################");
      analytics.logPageView(() => settings.label, route: settings.route, resolvedPath: settings.resolvedPath, params: {
        for (final entry in args.entries) entry.key: analyticsValue(entry.value),
      });
    } else {
      _log.warning("Wanted PathRouteSettings but found ${route.settings.runtimeType} for "
          "${route.settings.name ?? 'unnamed'} route $route.  Will not be logged");
    }
  }

  @override
  void didReplace({Route<dynamic>? newRoute, Route<dynamic>? oldRoute}) {}

  @override
  void didRemove(Route<dynamic> route, Route<dynamic>? previousRoute) {}

  @override
  void didPop(Route<dynamic> route, Route<dynamic>? previousRoute) {}
}

// Common place to send analytics events through.  Will dispatch to the appropriate analytics providers
class AnalyticsService implements IAnalyticsService {
  static final log = Logger("analyticsService");
  final List<AnalyticsProvider> providers;
  final _onReady = SafeCompleter<String>();
  String? anonymousId;
  Version? appVersion;
  final _userProps = <String, dynamic>{};
  final _globalProps = <String, dynamic>{};

  AnalyticsService._() : providers = defaultAnalyticsProviders;

  NavigatorObserver navObserver([String root = "/"]) {
    return AnalyticsNavigatorObserver(root);
  }

  Future initialize(DeviceInfo deviceInfo, AppInfo appInfo, UserPreferencesService prefs,
      {List<AnalyticsProvider>? provide}) async {
    // await FirebaseAnalytics().setAnalyticsCollectionEnabled(true);
    if (provide != null) {
      this.providers.clear();
      this.providers.addAll(provide);
    }
    anonymousId = await prefs.getString(analyticsIdKey);
    if (anonymousId == null) {
      anonymousId = deviceInfo.deviceId;
      await prefs.set(analyticsIdKey, anonymousId);
    }
    _globalProps.addAll(deviceInfo.toMap());
    _globalProps.addAll(appInfo.toMap());
    _onReady.complete(anonymousId);
  }

  @override
  void logEvent(dynamic name, [Map<String, dynamic>? parameters]) {
    _trackAll(name, parameters);
  }

  void encounteredError(Object error, [StackTrace? stackTrace, dynamic context]) {
    _onEach((p) => p.encounteredError(error, stackTrace, context));
  }

  void logLogout() {
    _trackAll("logout");
    _userProps.clear();
    _onEach((p) => p.reset(anonymousId));
  }

  void logShare({String? shareContentType, String? shareContentId, String? shareMethod, String? shareTarget}) {
    _onEach((p) => p.share(
          shareContentId: shareContentId,
          shareContentType: shareContentType,
          shareMethod: shareMethod,
          shareTarget: shareTarget,
        ));
  }

  void logLogin({String? userId, String? username}) {
    logEvent("loggedIn", {"userId": userId, "username": username});
    _onEach((p) => p.identify(userId: userId, username: username));
    _onEach((p) => p.setUserProperties(_globalProps));
  }

  Future logPageView(String? pageName(), {String? route, String? resolvedPath, Map<String, dynamic>? params}) async {
    _onEach((p) => p.logPageView(label: pageName(), route: route, params: params, resolvedPath: resolvedPath));
  }

  Future _onEach(void perform(AnalyticsProvider provider)) async {
    await _onReady.future;
    for (final p in providers) {
      try {
        perform(p);
      } catch (e) {
        log.warning("Error logging: $e", e);
      }
    }
  }

  void _trackAll(final eventName, [Map<String, dynamic>? properties]) {
    if (eventName == null) return;
    _onEach((p) => p.track(eventName, {
          ...?properties,
          ..._userProps,
        }));
  }

  void logAccountCreated({required String userId, required String username}) {
    identify(userId: userId, username: username);
    _trackAll(const ["account", "created"], {"method": "app", "userId": userId});
  }

  void setUserProperty(String? property, String? value) {
    if (property == null || value == null) return;
    _userProps[property] = value;
    _onEach((p) {
      p.setUserProperties({property: value});
    });
  }

  void logBeginCheckout({List products = const []}) {
    int i = 0;
    logEvent("beginCheckout", {
      for (final p in products.orEmpty()) "product${i++}": p?.toString() ?? 'Unknown product',
    });
  }

  void logPresentOffer({String? itemId, String? itemName, String? itemCategory, double? price, int? quantity}) {
    logEvent("presentOffer",
        {"itemId": itemId, "itemName": itemName, "itemCategory": itemCategory, "price": price, "quantity": quantity});
  }

  void identify({required String userId, String? username}) {
    _onEach((p) {
      p.identify(userId: userId, username: username);
      if (username != null) {
        p.setUserProperties({"username": username});
      }
    });
  }

  Future addUserInfo({String? name, String? email, String? phone}) async {
    if (email != null) {
      _onEach((provider) => provider.email = email).ignore();
    }
    if (name == null || name.toLowerCase() == "me") {
      name = email;
    }
    if (name != null) _onEach((provider) => provider.name = name).ignore();

    if (phone != null) {
      setUserProperty("Phone", "$phone");
    }
  }
}

class SunnyCrashlytics implements AnalyticsProvider {
  final FirebaseCrashlytics _crashlytics = FirebaseCrashlytics.instance;

  SunnyCrashlytics();

  void track(dynamic eventName, [Map<String, dynamic>? properties]) {
    //
  }

  void logPageView({String? route, String? resolvedPath, String? label, Map<String, dynamic>? params}) {
    // no-op
  }

  @override
  void share({String? shareContentType, String? shareContentId, String? shareMethod, String? shareTarget}) {}

  void identify({String? userId, String? username}) {
    _crashlytics.setUserIdentifier(userId!);
    if (username != null) {
      email = username;
    }
  }

  @override
  void encounteredError(Object error, [StackTrace? stackTrace, dynamic context]) {
    _crashlytics.recordError(error, stackTrace, reason: context);
  }

  @override
  set email(String email) {
    _crashlytics..setCustomKey("email", email);
  }

  @override
  set name(String? name) {
    if (name != null) {
      _crashlytics.setUserIdentifier(name);
    }
  }

  void setUserProperties(Map<String, dynamic> properties) {
    properties.forEach((key, value) {
      if (value is! Object) {
        return;
      } else {
        _crashlytics.setCustomKey(key, value);
      }
    });
  }

  void reset(String? anonymousId) => _crashlytics.setUserIdentifier(anonymousId!);
}
