import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:logging/logging.dart';
import 'package:pub_semver/pub_semver.dart' as semver;
import 'package:pub_semver/pub_semver.dart';
import 'package:sunny_lifecycle/app_info.dart';

final _log = Logger("appVersionService");

typedef GetMinAppVersion = Future<VersionCheckInfo> Function();
typedef AppInfoWidgetBuilder = Widget Function(
    BuildContext context, AppInfo info);

class VersionCheckSettings {
  final GetMinAppVersion getMinAppVersion;
  final AppInfoWidgetBuilder buildForcedUpgradeWidget;

  VersionCheckSettings(
      {required this.getMinAppVersion,
      required this.buildForcedUpgradeWidget});
}

abstract class VersionCheckInfo {
  Version? get forceUpgradePrior;

  Version? get warnUpgradePrior;

  const factory VersionCheckInfo(
      {required Version forceUpgradePrior,
      required Version warnUpgradePrior}) = _VersionCheckInfo;

  factory VersionCheckInfo.parsed(
          {required String forceUpgradePrior,
          required String warnUpgradePrior}) =>
      _VersionCheckInfo(
        forceUpgradePrior: semver.Version.parse(forceUpgradePrior),
        warnUpgradePrior: semver.Version.parse(warnUpgradePrior),
      );
}

class _VersionCheckInfo implements VersionCheckInfo {
  final Version? forceUpgradePrior;
  final Version? warnUpgradePrior;

  const _VersionCheckInfo({this.forceUpgradePrior, this.warnUpgradePrior});
}

VersionCheck checkAppVersion({
  required AppInfo appInfo,
  VersionCheckInfo? versionInfo,
}) {
  if (kIsWeb) return VersionCheck.noCheck(appInfo);
  //Get Current installed version of app

  final currentVersion = appInfo.appVersion;
  try {
    final forceVersion = versionInfo!.forceUpgradePrior!;
    final warnVersion = versionInfo.warnUpgradePrior;

    if (forceVersion > currentVersion) {
      return VersionCheck(
          appInfo: appInfo,
          result: VersionCheckResult.force,
          expectedVersion: forceVersion);
    } else if (warnVersion! > currentVersion) {
      return VersionCheck(
          appInfo: appInfo,
          result: VersionCheckResult.warn,
          expectedVersion: warnVersion);
    } else {
      return VersionCheck(
          appInfo: appInfo,
          result: VersionCheckResult.ok,
          expectedVersion: warnVersion);
    }
  } catch (exception) {
    _log.info('Unable to fetch remote config. Cached or default values will be '
        'used: $exception');

    return VersionCheck(appInfo: appInfo, result: VersionCheckResult.error);
  }
}

class VersionCheck extends Equatable {
  final AppInfo appInfo;
  final VersionCheckResult result;
  final Version? expectedVersion;

  VersionCheck(
      {required this.appInfo, required this.result, this.expectedVersion});

  bool get isOK => result == VersionCheckResult.ok;

  bool get isWarn => result == VersionCheckResult.warn;

  bool get isForce => result == VersionCheckResult.force;

  bool get isError => result == VersionCheckResult.error;

  Future<String?> get appStoreUrl async {
    return appInfo.appStoreUrl;
  }

  VersionCheck.noCheck(AppInfo appInfo)
      : appInfo = appInfo,
        result = VersionCheckResult.ok,
        expectedVersion = Version(0, 0, 0);

  @override
  List<Object?> get props => [appInfo, result, expectedVersion];

  @override
  String toString() {
    return "Result: ${result} -- expected=${expectedVersion}, actual=${appInfo.appVersion}";
  }
}

enum VersionCheckResult { ok, warn, force, error }
