import 'dart:async';

import 'package:firebase_auth/firebase_auth.dart' as fb;
import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_lifecycle/modules/firebase_auth_module.dart';
import 'package:sunny_lifecycle/sunny_lifecycle.dart';
import 'package:sunny_sdk_core/auth/firebase_api_auth.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';

const kPreconditionFailed = 412;
const kPreconditionRequired = 428;
const kUnauthorized = 401;
const kForbidden = 403;

class AuthEvent {
  final fb.User? fbUser;
  final dynamic error;
  final AuthEventSource source;

  const AuthEvent(this.fbUser, {required this.source}) : error = null;

  const AuthEvent.existing(this.fbUser,
      {this.source = AuthEventSource.existing})
      : error = null;

  const AuthEvent.error(this.error)
      : fbUser = null,
        source = AuthEventSource.error;
}

final _authStateClearinghouse = StreamController<AuthEvent>.broadcast();
final _unauthenticatedErrors = StreamController<bool>.broadcast();
final userStateChangedStream =
    SyncStream<AuthUserProfile>.controller(debugName: 'userStateChanged');

Stream<AuthUserProfile> get globalUserStateChanged =>
    userStateChangedStream.flatten(null, true).whereType();

typedef ProfileLoader = Future<UserDetails> Function(fb.User firebaseUser);
typedef ProfileName<P> = String Function(P profile);
typedef AccountCompleter<P extends UserDetails> = Future<P> Function(
    BuildContext context, fb.User? profile, String accessToken);

class FirebaseAuthState
    with LoggingMixin, LifecycleAwareMixin
    implements IAuthState {
  FirebaseAuthState(
      this.authModule, this._currentFbUser, UserDetails? currentProfile,
      {required Stream<ErrorStack> appErrors})
      : _loginProviders = authModule.loginProviders,
        _currentUserProfile = AuthUserProfile(
          _currentFbUser,
          currentProfile,
          AuthEventSource.initial,
        ) {
    /// This block receives all events from the auth state change clearing  house,
    /// loads the user profile if necessary, and pipes the results to the [onUserChange]
    /// stream
    authStateChange.asyncMap<AuthUserProfile>((authEvent) async {
      final fbUser = authEvent.fbUser;
      log.warning(
          "User changed: ${fbUser?.displayName ?? fbUser?.email ?? 'None'}");
      if (fbUser == null) {
        _currentUserProfile = AuthUserProfile.empty(authEvent.source);
        return _currentUserProfile!;
      } else {
        // Fetch the profile first thing.
        try {
          final profile = await authModule.loadProfile(fbUser);
          final aup = AuthUserProfile(fbUser, profile, authEvent.source);
          _currentUserProfile = aup;
          log.warning("User loaded from server: ${profile.name}");
          log.info("Profile ${profile.name} loaded");

          final token = await fbUser.getIdToken();
          final meta = fbUser.metadata;
          log.info("Last sign in: ${meta.lastSignInTime}");
          log.info("Token provider: ${fbUser.providerData}");
          log.info("Updated auth token: $token");
          return aup;
        } on ApiResponseException catch (e) {
          if (e.statusCode == kPreconditionFailed ||
              e.statusCode == kPreconditionRequired) {
            log.info(
                "Authenticated, but no account exists yet.  When the account is finished "
                "being created, you will have to manually push that userProfile through");
            return AuthUserProfile(fbUser, null, authEvent.source,
                status: AuthStatus.partial);
          } else if (e.statusCode == 401) {
            log.warning(
                "Got unauthorized getting profile. ${e.message} ? '(No message)' Logging you out!");
          }
          return AuthUserProfile.empty(authEvent.source);
        } catch (e, stack) {
          log.severe("Problems logging in: $e", e, stack);
          return AuthUserProfile.empty(AuthEventSource.failed);
        }
      }
    }).listen((mapped) {
      _currentUserProfile = mapped;
      userStateChangedStream.current = mapped;
    }, cancelOnError: false).watch(this);

    /// Reads from firebase auth change events, and pushes to clearing house
    fb.FirebaseAuth.instance.idTokenChanges().listen((event) async {
      log.info("Firebase sent auth state change: $event");
      if (_currentFbUser?.uid != event?.uid) {
        _currentFbUser = event;
        FirebaseApiAuth.user = event;
        _authStateClearinghouse
            .add(AuthEvent(event, source: AuthEventSource.framework));
      } else {
        log.info("Ignoring duplicate state change for ${event?.uid}");
      }
    }).watch(this);

    /// Listen for authentication failures and remove logged in users
    appErrors
        .where((_) {
          return _.exception is ApiResponseException;
        })
        .where((_) => (_.exception as ApiResponseException).statusCode == 401)
        .listen((_) {
          log.warning("Got unauthenticated error from API.");
          _authStateClearinghouse.add(AuthEvent.error(_));
          _unauthenticatedErrors.add(true);
        })
        .watch(this);

    Future.microtask(() {
      initializeState(fbUser: _currentFbUser, profile: _currentUserProfile);
    });
  }

  /// Most of the time, setting these variables should be handled automatically, and this method should not be needed.
  void initializeState(
      {required fb.User? fbUser,
      AuthUserProfile? profile,
      bool resetState = false}) {
    if (resetState) {
      // This will not fire any events, but will ensure that the next change is propagated.
      _currentUserProfile = null;
      _currentFbUser = null;
    }
    _authStateClearinghouse
        .add(AuthEvent(fbUser, source: AuthEventSource.initial));
    if (profile != null) {
      userStateChangedStream.current = profile;
    }
  }

  final FirebaseAuthAppModule authModule;

  List<LoginProvider> _loginProviders = [
    LoginProvider.google,
    LoginProvider.apple
  ];

  Stream<bool> get unauthenticatedErrors => _unauthenticatedErrors.stream;

  Stream<AuthEvent> get authStateChange =>
      Stream.value(AuthEvent.existing(_currentFbUser))
          .followedBy(_authStateClearinghouse.stream)
          .debounce(300.ms, leading: true);

  AuthUserProfile? _currentUserProfile;
  fb.User? _currentFbUser = fb.FirebaseAuth.instance.currentUser;

  UserDetails? get currentUser => _currentUserProfile?.profile;

  AuthUserProfile? get currentAuth => _currentUserProfile;

  set currentUser(UserDetails? currentUser) {
    userStateChangedStream.current =
        AuthUserProfile(_currentFbUser, currentUser, AuthEventSource.manual);
  }

  fb.User? get currentFirebaseUser => _currentFbUser;

  String? get accountId => currentAuth?.profile?.id;

  Stream<AuthUserProfile?> get onUserChange =>
      userStateChangedStream.flatten(null, false);

  Stream<AuthUserProfile?> get onUserChangeAfter =>
      userStateChangedStream.after;

  @override
  Future doShutdown() async {}

  @override
  bool operator ==(Object other) =>
      identical(this, other) || other is FirebaseAuthState;

  @override
  int get hashCode => 0;

  void onPreLogout(Future onPreLogout()) {
    if (!_preLogout.contains(onPreLogout)) {
      _preLogout.add(onPreLogout);
    }
  }

  void removePreLogout(Future onPreLogout()) {
    _preLogout.removeWhere((_) => _ == onPreLogout);
  }

  final List<AsyncCallback> _preLogout = [];

  Future signOut() async {
    for (var preLogoutHandler in _preLogout) {
      try {
        await preLogoutHandler();
      } catch (e, stack) {
        log.severe("Error running prelogout handler: $e", e, stack);
      }
    }

    /// Sign out with each provider, in case there is anything cached
    for (final provider in _loginProviders) {
      try {
        provider.signOut();
      } catch (e, stack) {
        log.severe("Error logging out with ${provider.name}", e, stack);
      }
    }

    /// Finally, sign out with Firebase
    await fb.FirebaseAuth.instance.signOut();
  }

  Future<bool> completeSignup(BuildContext context) async {
    assert(_currentFbUser != null);

    /// This will proceed with creating an account
    try {
      final token = await _currentFbUser!.getIdToken(true);
      final profile = await authModule.accountCompleter
          ?.call(context, _currentFbUser, token);

      _currentUserProfile =
          AuthUserProfile(_currentFbUser, profile, AuthEventSource.postSignup);
      userStateChangedStream.current = _currentUserProfile!;
      return true;
    } catch (e) {
      log.severe("Error completing signup: $e", e);

      return false;
    }
  }

  AuthUserProfile? get current => _currentUserProfile;

  @override
  Stream<AuthUserProfile> get userStateStream => globalUserStateChanged;

  @override
  bool get isLoggedIn => _currentFbUser != null;

  @override
  bool get isNotLoggedIn => !isLoggedIn;
}

extension AuthUserStreamExt<P> on Stream<AuthUserProfile> {
  Stream<UserDetails?> get profile => this.map((auth) => auth.profile);
}
