import 'dart:async';

import 'package:firebase_auth/firebase_auth.dart' as fb;
import 'package:firebase_core/firebase_core.dart';
import 'package:firebase_crashlytics/firebase_crashlytics.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:sunny_platform_widgets/sunny_platform_widgets.dart';
import 'package:logging/logging.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';
import 'package:responsive_builder/responsive_builder.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:sunny_lifecycle/analytics/analytics_service.dart';
import 'package:sunny_lifecycle/app_context_events.dart';
import 'package:sunny_lifecycle/app_info.dart';
import 'package:sunny_lifecycle/app_version_service.dart';
import 'package:sunny_lifecycle/auth_state.dart';
import 'package:sunny_lifecycle/forced_upgrade.dart';
import 'package:sunny_lifecycle/help/help_service.dart';
import 'package:sunny_lifecycle/login_provider.dart';
import 'package:sunny_lifecycle/sunny_app.dart';
import 'package:sunny_lifecycle/tabs.dart';
import 'package:sunny_lifecycle/user_preferences_service.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';

import 'lifecycle_extensions.dart';
import 'lifecycle_main_defaults.dart';
import 'modules/lifecycle_modules.dart';

final _mainLog = Logger("lifecycle");
final appErrorStream = StreamController<ErrorStack>.broadcast();
final rootNavigatorKey = GlobalKey<NavigatorState>(debugLabel: "rootNavigator");

Stream<ErrorStack> get appErrors => appErrorStream.stream;

typedef ChildWidgetBuilder = Widget Function(
    BuildContext context, Widget child);

typedef Themes ThemesBuilder(Brightness? brightness);

/// These are the variables loaded before we attempt to start the flutter runtime.
class LifecycleInit {
  /// The currently authenticated firebase user
  fb.User? get initFbUser => authState.currentUser?.source();

  /// The currently authenticated user details (from your system)
  UserDetails? get initProfile => authState.currentUser;

  final IAuthState authState;
  final UserPreferencesService userPrefs;
  final AnalyticsService analytics;
  final DeviceInfo deviceInfo;
  final RoutingModule? routing;
  final IHelpService? helpService;

  /// Your app's info, version, etc
  final AppInfo appInfo;

  LifecycleInit(
    this.deviceInfo,
    this.authState,
    this.analytics,
    this.routing,
    this.appInfo,
    this.userPrefs,
    this.helpService,
  );
}

/// This class accepts a wide array of strategies for various things, and then
/// loads the app in a way that makes sure everything is scoped properly.
///
/// It takes care of:
/// * App build versioning metadata
/// * Device/screen layouts
/// * User/authentication, displaying auth screens, logging out
/// * Push notifications (Firebase)
/// * App routing (using fluro)  This includes deep linking
/// * Analytics, crashlytics, error reporting
///
class AppBootstrap {
  static AppBootstrap? _current;

  BuildContext? _innerContext;

  BuildContextResolver get serviceResolver {
    return injection.serviceResolver;
  }

  set innerContext(BuildContext? context) {
    _innerContext = context;
    sunny.core.buildContext = context;
  }

  BuildContext? get innerContext => _innerContext;

  /// Stateful property that contains the parameters gathered during initialization
  LifecycleInit? lifecycleInit;

  final AppInfoModule appInfo;
  final AuthModule auth;
  final RoutingModule routing;
  final DeviceInfoModule deviceInfo;
  final UserPrefsModule userPrefs;
  final LocalizationModule localization;
  final AnalyticsModule analyticsInit;
  final PushNotificationModule? pushNotifications;
  final InjectionModule injection;

  static AppBootstrap current() {
    return _current ??
        illegalState("No app bootstrap has been invoked.  This "
            "cannot be invoked before the run method has been called");
  }

  static BuildContext currentContext() {
    return _current?.innerContext ??
        illegalState("No app bootstrap has been invoked.  This "
            "cannot be invoked before the run method has been called");
  }

  final bool isChangeBrightness;
  final String identifier;

  // final String googleApiKey;
  // final String googleSignInClientId;
  final String? baseApiUrl;

  final TabSet? tabs;

  /// You can provide this to enforce that your application is within a certain
  /// version range
  final VersionCheckSettings? versionCheckSettings;
  final Future Function(LifecycleInit? init)? completeStage1;
  final Widget Function(Widget inner) buildOuterContext;
  final ThemesBuilder themesBuilder;
  final void Function(BuildContext context) onUnauthenticatedError;
  final void Function(BuildContext context)? initWithinApp;
  final void Function(BuildContext context)? initWithinNavigator;

  final List<LoginProvider> loginProviders;

  final ScreenInit screenInit;
  final Widget Function(BuildContext context, Widget app) buildPlatformApp;
  final Stream<ErrorStack> appErrors;
  final Future<bool> Function()? onPreLogout;
  final FRouter? router;
  final FirebaseApp? firebaseApp;

  // final void Function(MetaFormService) metaFormServiceConfig;
  final String appTitle;
  final Widget Function(UserDetails profile)? loggedInWidget;
  final Getter<Widget>? loggedOutWidget;
  final Getter<Widget>? splashPageWidget;

  AppBootstrap(
      {this.auth = const AnonymousAuthModule(),
      required this.routing,
      this.appInfo = const DefaultAppInfoModule(),
      this.deviceInfo = const DefaultDeviceInfoModule(),
      this.userPrefs = const DefaultUserPrefsModule(),
      this.localization = const DefaultLocalizationModule(),
      this.analyticsInit = const DefaultAnalyticsModule(),
      this.pushNotifications,

      /// Plugs into [Sunny] to handle service resolution.  This allows for a
      /// swappable strategy that may use `provider` or `kiwi`.  By default,
      /// this uses the Provider package
      this.injection = const ProviderInjectionModule(),
      this.tabs,

      /// Set this if you want to check if this version of the app is too old
      this.versionCheckSettings,

      // /// API key to be used for making calls to google
      // @required this.googleApiKey,
      //
      // /// API key to be used for making calls to google
      // @required this.googleSignInClientId,

      /// Whether the app should attempt to change brightness
      this.isChangeBrightness = true,

      /// The initialized firebase app
      this.firebaseApp,

      /// Executes code at the end of stage1, before anything Flutter related
      /// loads.  This gives a chance to do any pre-work before Widgets start flying.
      ///
      /// If the return value is a [SingleChildWidget] ([Provider]) or variant, it will
      /// be added to the top-level multi-provider
      this.completeStage1 = defaultCompleteStage1,

      /// App identifier.  Used to prefix preferences, among other things
      required this.identifier,

      /// Builds the innerMost widget, wrapping the PlatformApp
      this.buildPlatformApp = defaultBuildPlatformApp,

      /// The fluro router to use for path resolution
      this.router,

      /// Executes once within the PlatformApp component
      this.initWithinApp,

      /// Executes once within the PlatformApp component
      this.initWithinNavigator,

      /// Executes when an unauthenticated exception occurs, allows you
      /// to swap for a login screen
      this.onUnauthenticatedError = noopBuildContext,

      /// The widget to display when logged in
      this.loggedInWidget,

      /// The widget to display when logged out
      this.loggedOutWidget,

      /// The widget to display when logged out
      this.splashPageWidget,

      /// The title of the app
      required this.appTitle,

      /// The base API url for this app
      this.baseApiUrl,

      /// Builds the outer widget
      this.buildOuterContext = defaultBuildOuterContext,

      /// Builds the app themes, based on brightness
      this.themesBuilder = buildDefaultThemes,

      /// Login providers to use
      this.loginProviders = const [],

      /// A stream of errors that bubbled up to the top level
      Stream<ErrorStack>? appErrors,

      /// Executed when the screen changes sizes
      this.screenInit = defaultScreenInit,

      /// Runs right before the user logs out.  Useful for doing any cleanup
      /// before cleaning up state
      this.onPreLogout})
      : appErrors = appErrors ?? Stream.empty();

  Future<Widget> buildRootWidget() async {
    AppBootstrap._current = this;

    /// Set the resolver
    final serviceResolver = await injection.initializeResolver();
    final infoX = await localization.initializeInfoX();

    final _appInfo = await appInfo.loadAppInfo(identifier, appTitle);

    _mainLog.info("---------------------------------------- ");
    _mainLog.info("---------------------------------------- ");
    _mainLog.info("");
    _mainLog.info("${_appInfo.appName}");
    _mainLog.info("Package: ${_appInfo.packageName}");
    _mainLog.info("Version: ${_appInfo.appVersion}");
    if (_appInfo.buildNumber.isNotNullOrBlank) {
      _mainLog.info("Package: ${_appInfo.buildNumber}");
    }
    _mainLog.info("");
    _mainLog.info("---------------------------------------- ");
    _mainLog.info("---------------------------------------- ");
    var initProfile = await auth.getInitialUser();

    _mainLog.info("");
    _mainLog.info("Starting user: ${initProfile?.profile?.name ?? 'None'}");

    await routing.initializeRoutes();

    _mainLog.info("Fetching device info...");
    var _deviceInfo = await deviceInfo.loadDeviceInfo();
    _mainLog.info("Fetched device: ${_deviceInfo.deviceType}");

    final _userPrefs = userPrefs.initializePreferences(identifier);
    final analytics = analyticsInit.initializeAnalytics();
    try {
      await analytics.initialize(
        _deviceInfo,
        _appInfo,
        _userPrefs,
      );
      _mainLog.info("Initialized Firebase Analytics");
    } catch (e) {
      _mainLog.warning("Unable to initialize Firebase analytics: $e");
    }

    final authState = auth.initializeState(initProfile, appErrors);

    if (onPreLogout != null) {
      authState.onPreLogout(onPreLogout!);
    }

    this.lifecycleInit = LifecycleInit(
      _deviceInfo,
      authState,
      analytics,
      routing,
      _appInfo,
      _userPrefs,
      null,
    );

    var appContextEvents = AppContextEvents();

    /// Register app events stream

    if (initWithinNavigator != null) {
      appContextEvents.doWithAppContext(initWithinNavigator!);
    }
    final res = await completeStage1?.call(lifecycleInit);

    Widget? builtApp;

    final sunnyLocations = await localization.initializeLocalization(infoX);

    return Builder(
      builder: (context) {
        if (builtApp == null) {
          serviceResolver.registerSingleton<AppContextEvents>(
              context, appContextEvents);
          serviceResolver.registerSingleton<BuildContextResolver>(
              context, serviceResolver);
        }
        return builtApp ??= serviceResolver.register(
          context,
          [
            Inst.constant(infoX),
            Inst.constant(sunnyLocations),
          ],
          child: MultiProvider(
            key: Key("mainTop"),
            providers: [
              Provider.value(
                value: appContextEvents,
                updateShouldNotify: never,
              ),

              Provider.value(value: infoX, updateShouldNotify: never),
              Provider.value(value: sunnyLocations, updateShouldNotify: never),
              Provider.value(value: this, updateShouldNotify: never),
              Provider.value(value: _deviceInfo, updateShouldNotify: never),
              Provider.value(value: _appInfo, updateShouldNotify: never),
              Provider.value(value: _userPrefs, updateShouldNotify: never),
              Provider.value(value: lifecycleInit, updateShouldNotify: never),
              // Provider.value(value: help, updateShouldNotify: never),
              Provider.value(value: router, updateShouldNotify: never),
              Provider.value(
                  value: lifecycleInit!.authState, updateShouldNotify: never),
              Provider<IAuthState>.value(
                  value: lifecycleInit!.authState, updateShouldNotify: never),
              if (res is SingleChildWidget) res,
            ],
            child: buildOuterContext(
              AppWidgets(
                deviceInfo: _deviceInfo,
                appInfo: _appInfo,
                isChangeBrightness: isChangeBrightness,
              ),
            ),
          ),
        );
      },
    );
  }

  List<Object> get modules => [
        appInfo,
        auth,
        routing,
        deviceInfo,
        userPrefs,
        localization,
        analyticsInit,
        if (pushNotifications != null) pushNotifications!,
        injection,
      ];

  void initializeFromWithin(BuildContext innerContext) {
    this.innerContext = innerContext;
    initWithinApp?.call(innerContext);
  }

  Future run() async {
    AppBootstrap._current = this;

    if (!kIsWeb) {
      FirebaseCrashlytics.instance
          .setCrashlyticsCollectionEnabled(!kIsWeb && !kDebugMode);
    }
    // Pass all uncaught errors from the framework to Crashlytics.
    if (!kIsWeb) {
      FlutterError.onError = FirebaseCrashlytics.instance.recordFlutterError;
    }

    runZonedGuarded(() async {
      runApp(await buildRootWidget());
    }, (obj, stackTrace) {
      appErrorStream.add(ErrorStack(obj, stackTrace));
      _mainLog.severe("Error in app: $obj", obj, stackTrace);

      analytics.encounteredError(obj, stackTrace, "App catchall");
    });
  }
}

Widget buildDefaultPlatformApp(BuildContext context, Widget app) {
  return app;
}

Themes buildDefaultThemes(Brightness? brightness) => Themes(
    brightness: brightness ?? Brightness.light,
    cardTheme: PlatformCardTheme(),
    cupertinoTheme: CupertinoThemeData(),
    darkTheme: ThemeData(),
    lightTheme: ThemeData(),
    themeBuilder: (_, __, ___) => ThemeData(),
    visualStyle: CupertinoVisualStyle(),
    listTileTheme: ListTileTheme(child: Container()));

bool notEquals<T>(T a, T b) => a != b;

bool whenNotNull<T>(T a, T b) => a == null && b != null;

bool never<T>(T a, T b) => false;

class AppWidgets extends StatefulWidget {
  final AppInfo appInfo;
  final bool isChangeBrightness;
  final DeviceInfo? deviceInfo;

  const AppWidgets({
    Key? key,
    required this.deviceInfo,
    required this.appInfo,
    this.isChangeBrightness = true,
  }) : super(key: key);

  @override
  _AppWidgetsState createState() => _AppWidgetsState();
}

typedef Future<T?> DeepLinkHandler<T>(BuildContext context);

class _AppWidgetsState extends BaseState<AppWidgets>
    with WidgetsBindingObserver {
  late LayoutInfo _layoutInfo;
  // ignore: unused_field
  PlatformStyle? _targetPlatform;
  late Size _screenSize;
  DeviceInfo? _deviceInfo;
  Brightness? _brightness;
  AccountCompleter? signupCompleter;

  DateTime? _lastVersionNag;
  bool _forceVersion = false;

  int rebuildNum = 1;

  @override
  void didChangePlatformBrightness() {
    final brightness = WidgetsBinding.instance!.window.platformBrightness;
    if (brightness != _brightness) {
      _mainLog.info("Brightness changed: from $_brightness to $brightness");
      setState(() {
        this._updateBrightness(brightness);
      });
    }
  }

  _updateBrightness(Brightness b) {
    this._brightness = b;
    SunnyColors.resolveBrightness(context);
    // rebuildNum++;
  }

  @override
  void didChangeMetrics() {
    final size = WidgetsBinding.instance!.window.physicalSize /
        WidgetsBinding.instance!.window.devicePixelRatio;

    if (size != _layoutInfo.screenSize) {
      setState(() {
        _screenSize = size;
        _resync();
        // rebuildNum++;
      });
    }
  }

  void _resync() {
    final screenType = getDeviceType(_screenSize);
    final _target = _calculatePlatformStyle(screenType);
    _layoutInfo = LayoutInfo.ofSize(
        _screenSize, screenType, _target.first, _target.second);

    _mainLog.info("Resynced platform info: ");
    _mainLog.info("  - style: ${_layoutInfo.platformStyle}");
    _mainLog.info("  - target: ${_layoutInfo.targetPlatform}");
    _mainLog.info("  - size: ${_layoutInfo.screenSize}");
    _mainLog.info("  - screenType: ${_layoutInfo.screenType}");
  }

  @override
  void initState() {
    super.initState();
    _deviceInfo = widget.deviceInfo ?? context.get();
    assert(_deviceInfo != null);
    WidgetsBinding.instance!.addObserver(this);
    _screenSize = WidgetsBinding.instance!.window.physicalSize /
        WidgetsBinding.instance!.window.devicePixelRatio;
    this._brightness = WidgetsBinding.instance!.window.platformBrightness;
    _resync();
    _checkAppVersion();
    _registerNavContext(context);
  }

  Tuple<PlatformStyle, TargetPlatform> _calculatePlatformStyle(
      DeviceScreenType type) {
    final platform = defaultTargetPlatform;
    if (kIsWeb) {
      /// In the case of web, we still want our app to respond accordingly
      final deviceBrand = _deviceInfo?.deviceBrand?.toLowerCase();
      if (deviceBrand == "apple") {
        switch (type) {
          // ignore: deprecated_member_use
          case DeviceScreenType.Desktop:
          case DeviceScreenType.desktop:
            return PlatformStyle.Material.to(TargetPlatform.macOS);
          default:
            return PlatformStyle.Cupertino.to(TargetPlatform.iOS);
        }
      } else {
        return PlatformStyle.Material.to(TargetPlatform.macOS);
      }
    }
    switch (platform) {
      case TargetPlatform.android:
        return PlatformStyle.Material.to(platform);
      case TargetPlatform.fuchsia:
        return PlatformStyle.Material.to(platform);
      case TargetPlatform.iOS:
        return PlatformStyle.Cupertino.to(platform);
      case TargetPlatform.linux:
        return PlatformStyle.Material.to(platform);
      case TargetPlatform.macOS:
        return PlatformStyle.Cupertino.to(platform);
      case TargetPlatform.windows:
        return PlatformStyle.Material.to(platform);
    }
  }

  void dispose() {
    WidgetsBinding.instance!.removeObserver(this);
    super.dispose();
  }

  Future _checkAppVersion() async {
    var current = AppBootstrap.current();
    final versionChecking = current.versionCheckSettings;
    final appInfo = current.lifecycleInit!.appInfo;
    if (versionChecking != null) {
      final versionInfo = await versionChecking.getMinAppVersion();
      try {
        var check = checkAppVersion(
          appInfo: appInfo,
          versionInfo: versionInfo,
        );
        _mainLog.info("Result of version check: $check");
        switch (check.result) {
          case VersionCheckResult.force:
            setState(() {
              _forceVersion = true;
            });
            return;
          case VersionCheckResult.warn:
            if (_lastVersionNag == null || _lastVersionNag!.daysAgo > 5) {
              _lastVersionNag = DateTime.now();
              _doVersionNag(appInfo);
            }
            return;
          default:
            break;
        }
      } catch (e) {
        _mainLog.warning("Unable to check version info $e", e);
      }
    }
  }

  void _doVersionNag(AppInfo appInfo) async {
    appContextEvents.doWithAppContext((context) async {
      return await showPlatformDialog<String>(
        context: context,
        barrierDismissible: false,
        builder: (BuildContext context) {
          final title = "New Update Available";
          final message =
              "There is a newer version of ${appInfo.appName} available.  We recommend that you update to the latest version.";
          final btnLabel = "Update Now";
          final btnLabelCancel = "Update Later";

          return PlatformAlertDialog(
            title: Text(title),
            content: Text(message),
            actions: [
              PlatformDialogAction(
                child: Text(btnLabel),
                onPressed: () => goToAppStore(context, appInfo),
              ),
              PlatformDialogAction(
                child: Text(btnLabelCancel),
                onPressed: () => Navigator.pop(context),
              ),
            ],
          );
        },
      );
    });
  }

  /// Whenever the app changes state, we should re-check the version status
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (state == AppLifecycleState.resumed) {
      _checkAppVersion();
    }
  }

  @override
  Widget build(BuildContext context) {
    var app = AppBootstrap.current();
    app.screenInit.call(context, _layoutInfo);
    final themes = context.app().themesBuilder(
        widget.isChangeBrightness ? _brightness : Brightness.light);

    if (_forceVersion == true) {
      return app.versionCheckSettings!
          .buildForcedUpgradeWidget(context, app.lifecycleInit!.appInfo);
    } else {
      return app.serviceResolver.register(
        context,
        [
          Inst.constant(themes.cardTheme, skipIfRegistered: true),
          Inst<LayoutInfo>.instance(_layoutInfo),
          Inst<Brightness>.instance(
              widget.isChangeBrightness ? _brightness : Brightness.light),
          Inst.instance(themes.cupertinoTheme, skipIfRegistered: true),
          Inst<Themes>.instance(
            themes,
            shouldNotify: (a, b) => a.brightness != b.brightness,
          ),
        ],
        key: Key("top-providers"),
        child: Material(
          key: Key('build-$rebuildNum'),
          child: Theme(
            data: _brightness == Brightness.dark
                ? themes.darkTheme
                : themes.lightTheme,
            child: CupertinoTheme(
              data: themes.cupertinoTheme,
              child: themes.buildListTileTheme(
                child: SunnyApp(widget.appInfo, app: app),
              ),
            ),
          ),
        ),
      );
    }
  }

  /// Registers initializers that require a context that has access to
  /// the navigator
  void _registerNavContext(BuildContext outerContext) {
    AppContextEvents.of(outerContext).doWithAppContext((context) async {
      AppBootstrap.current().initWithinNavigator?.call(context);
    });
  }
}

// class SunnyRootPage extends StatefulWidget {
//   const SunnyRootPage({Key key}) : super(key: key);
//
//   @override
//   _SunnyRootPageState createState() => _SunnyRootPageState();
// }
//
// class _SunnyRootPageState extends StateObserver<SunnyRootPage>
//     with LoggingMixin {
//   bool get isLoggedIn => user != null;
//   dynamic error;
//   fb.User user;
//   StreamSubscription sub;
//   AppBootstrap _app;
//
//   /// If a deep link comes in and the user hasn't logged in, it gets captured here until
//   /// they do log in; then attempts to resolve
//   DeepLinkHandler _pendingRoute;
//
//   Stream<UserDetails> _stream;
//
//   AuthState _trippiAuth;
//
//   @override
//   void initState() {
//     super.initState();
//
//     _app = context.app();
//     _app.innerContext = context;
//     _trippiAuth = context.authState;
//     _stream = _trippiAuth.onUserChange.profile;
//
//     _trippiAuth.onUserChange.profile.listen((user) {
//       /// Applying to analytics
//       if (user != null) {
//         analytics.identify(userId: user.id, username: user.email);
//         analytics.addUserInfo(
//           name: user.name,
//           email: user.email,
//           phone: user.phone,
//         );
//       } else {
//         analytics.logLogout();
//       }
//     }).watch(this);
//
//     getStartingUrl().sanitized.nonNull((initUrl) {
//       _mainLog.info("Got app launch link: $initUrl.");
//       if (!kIsWeb) {
//         // not sure why, but on web, these links are followed automatically
//         final _pendingRoute = resolveDeepLinks(context, initUrl.toString());
//         if (_pendingRoute != null) {
//           _mainLog.info("   - $initUrl.  Resolved (${_pendingRoute != null})");
//           _pendingRoute(context);
//         } else {
//           _mainLog.warning("   - $initUrl. Could not be resolved");
//         }
//       }
//     });
//
//     getUriChangeStream().listen((link) async {
//       final sane = link.sanitized;
//       _mainLog.info("Got link change browsing: $link: sanitized to $sane");
//
//       if (!kIsWeb) {
//         final _pendingRoute = resolveDeepLinks(context, sane);
//         if (_pendingRoute != null) {
//           _mainLog.info("   - $sane.  Resolved (${_pendingRoute != null})");
//           _pendingRoute(context);
//         } else {
//           _mainLog.warning("   - $sane. Could not be resolved");
//         }
//       }
//     })?.watch(this);
//
//     /// This needs to happen far enough down the tree where there's a navigator.
//     _trippiAuth.unauthenticatedErrors.listen((_) {
//       _app.onUnauthenticatedError?.call(context);
//     }).watch(this);
//
//     _app.initWithinApp?.call(context);
//   }
//
//   @override
//   void dispose() {
//     super.dispose();
//   }
//
//   @override
//   Widget build(BuildContext context) {
//     return StreamBuilder<UserDetails>(
//       key: Key("userChanges"),
//       stream: _stream,
//       initialData: _trippiAuth.currentUser,
//       builder: (context, snap) =>
//           snap.render(context, allowNull: true, successFn: (profile) {
//         _mainLog.info("Building with ${profile?.id ?? 'no user'}");
//         return AnimatedSwitcher(
//           key: Key("loginSwitcher"),
//           child: profile == null
//               ? _app.loggedOutWidget()
//               : _app.loggedInWidget(profile),
//           duration: 400.ms,
//           switchInCurve: Curves.easeIn,
//           switchOutCurve: Curves.easeIn,
//         );
//       }, loadingFn: () {
//         return Container(color: Colors.transparent);
//       }),
//     );
//   }
// }

DeepLinkHandler? resolveDeepLinks(BuildContext context, uri) {
  final router = SunnyRouting.router;
  final match = router.matchRoute("$uri");
  if (match == null) {
    return null;
  } else {
    return (context) async {
      final result = await match.route.go(context, args: match.parameters);

      /// Inspect result?
      return result;
    };
  }
}

Locale get enUs {
  return Locale.fromSubtags(languageCode: 'en', countryCode: 'us');
}

// ignore: non_constant_identifier_names
void routeGo() {}
