import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:sunny_platform_widgets/sunny_platform_widgets.dart';
import 'package:sunny_core_widgets/core_ext.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_essentials/container/spaced.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:sunny_lifecycle/routing/route_history.dart';
import 'package:sunny_lifecycle/sunny_lifecycle.dart';

class FRouterDelegate extends RouterDelegate<AppRouteMatch>
    with
        PopNavigatorRouterDelegateMixin<AppRouteMatch>,
        ChangeNotifier,
        LoggingMixin,
        ThrottledChangeNotifier {
  final FRouter routes;
  final FRouteState state;
  final Widget Function(UserDetails details) loggedInWidget;
  final Widget Function()? loggedOutWidget;
  final Widget Function()? splashPageWidget;
  final bool allowAnonymousLogin;
  final HeroController heroController = HeroController();

  /// Optional property - child routers can provide tabs to improve routing
  /// logic.
  TabSet _tabs = TabSet.empty();
  final Map<String, int?> _tabPaths = {};

  /// Optional property - child routers can provide tabs to improve routing
  /// logic.
  TabSet get tabs => _tabs;

  set tabState(TabSet tabState) {
    log.info("Received ${tabState.length} tabs");
    state.tabState = tabState;
    _tabs = tabState;
    _tabPaths.clear();
    _tabs.forEach((item) {
      _tabPaths[item.route.route] = item.tabIndex;
    });
  }

  @override
  final Limiter limiter = Debouncer(delay: 300.ms, atBegin: true);

  @override
  final GlobalKey<NavigatorState> navigatorKey;

  FRouterDelegate(this.state, this.routes,
      // this.routeProvider,
      {required this.loggedInWidget,
      TabSet? tabs,
      this.splashPageWidget,
      this.loggedOutWidget,
      required this.allowAnonymousLogin,
      required this.navigatorKey}) {
    if (tabs != null) {
      tabState = tabs;
    }
    // _observer.addListener(_onRouterChanges);
    /// Listens to FRouteState state changes, notifies any router listeners,
    /// which will also trigger a rebuild so downstream widgets can adjust
    /// accordingly
    state.addListener(_onStateChange);
  }

  /// Called when state changes, basically, the change is propagated
  void _onStateChange() {
    /// Update

    log.info("NOTIFYING OF STATE CHANGE!");
    notifyWithLimit();
  }

  /// Listens to the browser url for any changes.
  // void _onRouteStateChange() {
  //   var loc = routeProvider.value?.location;
  //   _checkLocationForTabChange(loc);
  // }

  /// Determine if a location change (uri) should cause the selected tab
  /// to change also - This only happens when loading routes when the app
  /// is first loaded.
  void _checkLocationForTabChange(String? loc) {
    log.info("${loc ?? 'none'} (start)");

    if (loc != null) {
      var newLocationJsonPath = JsonPath.of(loc);
      if (rootKeys.contains(loc)) {
        log.info("  -> $loc [root=true]");
        return;
      }

      /// We should have returned by this point if the route was
      /// fully handled.  We're going to check for a matched route
      /// and queue them up to be pushed on top of the stack
      var _path = JsonPath.root();
      var _added = 0;
      int? _newTab;
      var _pending = [...state.rootRoutes];
      late RouteHistoryItem<int> history;
      for (var segment in newLocationJsonPath.segments) {
        _path += segment;
        if (_tabPaths.containsKey("$_path")) {
          /// A direct path to the tab.
          final index = _tabPaths["$_path"];
          log.info("  -> ${_path} [tab=$index] this route belongs to a tab");
          _newTab = index;
          history = RouteHistoryItem("$_path", index, isPage: true);
        } else if (_newTab != null ||
            _path.startsWith(JsonPath.of("/tribes"))) {
          // If _newTab != null, then we know this route belongs to a tab already
          // and we'll want to push each route along the way
          final matched = routes.matchRoute("$_path");

          if (matched != null) {
            log.info("  -> ${matched.resolvedUri} [push=true]}");
            var pending = PendingRoute(matched);
            if (!_pending.contains(pending)) {
              _added++;
              _pending *= (pending);
            }
            history =
                RouteHistoryItem(loc, _newTab ?? state.currTab, isPage: true);
          } else {
            log.info("  -> ${_path} [ignored=true]");
          }
        }
      }

      if (_newTab != null) {
        state.currTab = _newTab;
      }

      // This method is called when a route is parsed, outside of a normal
      // push, so we only need to track the history? This is a bad idea.
      state.pushHistoryItem(history);

      if (_added > 0) {
        log.info("  queueing up ${_added} routes (total ${_pending.length}");
        state.rootRoutes = _pending;
      }
    } else {
      log.info("  no location");
    }
  }

  /// Called by the [Router] when the [Router.routeInformationProvider] reports that a
  /// new route has been pushed to the application by the operating system.
  ///
  /// Consider using a [SynchronousFuture] if the result can be computed
  /// synchronously, so that the [Router] does not need to wait for the next
  /// microtask to schedule a build.
  @override
  Future<void> setNewRoutePath(AppRouteMatch configuration) {
    if (configuration == this.currentConfiguration) {
      return SynchronousFuture(null);
    }
    var resolvedUri = configuration.resolvedUri;

    log.info("new route: ${configuration.resolvedUri}");
    _checkLocationForTabChange(resolvedUri);
    notifyListeners();
    return SynchronousFuture(null);
  }

  @override
  void dispose() {
    log.info("Disposed routerDelegate");
    // _observer.dispose();
    state.removeListener(_onStateChange);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    var isLoggedIn = state.loggedIn?.profile != null &&
        (state.loggedIn?.fbUser?.isAnonymous == false || allowAnonymousLogin);
    log.info("Building LOGGED IN: $isLoggedIn");
    var pages = [
      if (isLoggedIn) ...[
        platformPage(
          context: context,
          key: appValueKey,
          name: appValueKey.value,
          child: NavigatorContextWidget(
              child: loggedInWidget(state.loggedIn!.profile!)),
        ),
        for (var p in state.rootRoutes.pages)
          platformPage(
            context: context,
            name: p.route,
            child: p.match!.route
                .asPageRoute()!
                .handleAny(context, p.match!.parameters),
            key: ValueKey("pending-${p.route}"),
          ),
      ] else if (state.isInitted != true)
        platformPage(
          context: context,
          key: splashValueKey,
          name: splashValueKey.value,
          child: NavigatorContextWidget(
              child: splashPageWidget?.call() ?? emptyBox),
        )
      else
        platformPage(
          context: context,
          key: loginValueKey,
          child: NavigatorContextWidget(
              child: loggedOutWidget?.call() ?? emptyBox),
          name: loginValueKey.value,
        ),
    ];
    log.info(
        "routes.build[pages=${pages.map((e) => e.name ?? e.key).join(";")}");
    return Navigator(
      key: navigatorKey,

      onPopPage: _handleRootNavigtorPopPage,
      observers: [heroController, RootNavigatorObserver(this)],
      onGenerateRoute: routes.onGenerateRoute,
      reportsRouteUpdateToEngine: false,
      // restorationScopeId: _restorationId,
      pages: pages,
    );
  }

  @override
  Future<bool> popRoute() async {
    if (state.currTab != null) {
      final res =
          await state.currentTabState?.navigatorKey.currentState?.maybePop();
      if (res == true) {
        _updateLocation();
        return true;
      }
    }
    log.info("Got pop route (but couldn't handle up) (delegating)");
    return super.popRoute().then((res) {
      if (res) {
        state.popRootRoute();
        _updateLocation();
      }
      return res;
    });
  }

  bool _handleRootNavigtorPopPage(Route<dynamic> route, dynamic res) {
    try {
      log.info("Popped $route (with $res)");

      final bool result = route.didPop(res);
      if (state.rootRoutes.isNotEmpty) {
        log.info(
            "  ^pending Remove pending ${state.rootRoutes.last.route}; didPop=$result");
        state.popRootRoute();
        _updateLocation();
        return true;
      } else {
        log.info("  !pending Return false; didPop=$result");
        return false;
      }
    } catch (e, stack) {
      log.severe("Error popping: $e", e, stack);
      return false;
    }
  }

  NavigatorState? get tabNavigator {
    return tabs[state.currTab!].navigatorKey.currentState;
  }

  String? get currentRoute {
    return state.rootRoutes.firstOrNull?.route;
  }

  @override
  String toString() {
    return 'FRouterDelegate{state: $state}';
  }

  @override
  AppRouteMatch? get currentConfiguration {
    // if (_pendingRoutes.isNotEmpty) {
    //   return _pendingRoutes.last.match;
    // } else if (state.currTab != null) {
    //   return AppRouteMatch(_tabs[state.currTab].route, null);
    // } else {
    //   return rootAppRouteMatch(AppBootstrap.current().splashPageWidget);
    // }
    return null;
  }

  void _updateLocation() {
    if (state.rootRoutes.isNotEmpty) {
      final newRoute = currentRoute;
      if (newRoute != null) {
        log.info("  @overlay: $newRoute Updated browser uri");
        SystemNavigator.routeInformationUpdated(location: newRoute);
      }
    } else if (state.currentTabRoutes.pages.isNotEmpty) {
      log.info(
          "  @tab: ${state.currentTabRoutes.pages.last.route} Updated browser uri");
      SystemNavigator.routeInformationUpdated(
          location: state.currentTabRoutes.pages.last.route);
    } else if (state.currTab != null) {
      log.info(
          "  @base: ${tabs[state.currTab!].route.route} Updated browser uri");
      SystemNavigator.routeInformationUpdated(
        location: tabs[state.currTab].route.route,
      );
    }
  }

  void reportTabPop(RouteInformation? info, {int? currTab}) {
    if (info == null) {
      return;
    }
    final matchRoute = fromRouteInformation(info);
    if (matchRoute == null) {
      state.popAnonymousRoute();
      log.info("  !tab-route ${info.location} (no pop)");
    } else {
      log.info("  ^tab-route ${info.location}");
      state.popChildRoute(state.currTab);
      _updateLocation();
    }
  }

  /// Since this nav
  void reportTabPush(RouteInformation? info, {int? currTab}) {
    if (info == null) {
      return;
    }
    var matchRoute = routes.matchAppRouteExt(info);

    var resolvedUri = matchRoute?.resolvedUri;
    if (resolvedUri == null) {
      state.pushAnonymousRoute(info.location ?? '/${uuid()}', currTab: currTab);
      return;
    }
    state.pushCurrentTabRoute(PendingRoute(matchRoute));
    _updateLocation();
  }

  void reportRootPop(RouteInformation? info) {
    if (info == null) {
      return;
    }
    final matchRoute = fromRouteInformation(info);
    if (matchRoute == null) {
      log.info("  !overlay ${info.location} (no pop)");
    } else {
      log.info("  ^overlay ${info.location}");
      state.popRootRoute();
      _updateLocation();
    }
  }

  void reportRootPush(RouteInformation? info) {
    if (info == null) {
      return;
    }

    final matchRoute = fromRouteInformation(info);
    final res = state.pushRootRoute(matchRoute, isPage: false);

    if (res) {
      log.info("  -> overlay ${info.location}");
      _updateLocation();
    } else {
      log.info("  ^ overlay ${info.location} [ignore=true]");
    }
  }

  AppRouteMatch? fromRouteInformation(RouteInformation info) {
    var matchRoute = routes.matchAppRouteExt(info);
    if (matchRoute?.resolvedUri == null) return null;
    return matchRoute;
  }
}

extension FRouterMatchExt on FRouter {
  AppRouteMatch? matchAppRouteExt(final RouteInformation? info) {
    if (info != null) {
      var matchRoute = info is PathRouteSettings
          ? AppRouteMatch(this.findRouteByUriTemplate(info.route),
              info.routeParams ?? DefaultRouteParams.empty)
          : this.matchRoute(info.location!);

      return matchRoute;
    }
    return null;
  }
}
