import 'package:firebase_core/firebase_core.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:logging/logging.dart';
import 'package:logging_config/logging_config.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:sunny_lifecycle/lifecycle_main.dart';
import 'package:sunny_lifecycle/modules/testing_modules.dart';
import 'package:sunny_lifecycle/routing/route_history.dart';
import 'package:sunny_lifecycle/routing2/fluro_routing_system.dart';

import 'mock_firebase.dart';
import 'test_routing_widgets.dart';

final tabNavigatorKey = GlobalKey<NavigatorState>();

void main() {
  group("Router tests - root", () {
    // TestWidgetsFlutterBinding.ensureInitialized(); Gets called in setupFirebaseAuthMocks()
    setupFirebaseAuthMocks();

    setUpAll(() async {
      await Firebase.initializeApp();
    });

    testWidgets("Route initializes to root", (WidgetTester tester) async {
      var router = createMainRouter();

      try {
        var history = TestCaptureHistoryObserver();

        var testAppWidget = await buildTestApp(router: router, historyObservers: [history]);
        await tester.pumpWidget(testAppWidget);
      } catch (e) {
        throw e;
      }

      expect(find.widgetWithText(RouteTestPage, 'Page 1 Page Body'), findsOneWidget);
    });

    testWidgets("Non-nested route change", (WidgetTester tester) async {
      var router = createMainRouter();
      var history = TestCaptureHistoryObserver();

      var testAppWidget = await buildTestApp(router: router, historyObservers: [history]);
      await tester.pumpWidget(testAppWidget);

      var switchPage2 = find.widgetWithText(MaterialButton, 'Switch to Page 2');
      expect(switchPage2, findsOneWidget);
      await tester.tap(switchPage2);
      await tester.pumpAndSettle();

      expect(find.widgetWithText(RouteTestPage, 'Page 2 Page Body'), findsOneWidget,
          reason: 'Tapping the button switched routes to page 2');

      await tester.tap(find.widgetWithText(MaterialButton, 'Switch to Page 3'));
      await tester.pumpAndSettle();

      expect(find.widgetWithText(RouteTestPage, 'Page 3 Page Body'), findsOneWidget,
          reason: 'Tapping the button switched routes to page 3');
    });
  });
}

Future<Widget> buildTestApp(
    {required FRouter router, String? initialPage, List<HistoryObserver> historyObservers = const []}) async {
  configureLogging(LogConfig.root(Level.INFO, handler: LoggingHandler.dev()));

  // var firebase = await Firebase.initializeApp();

  final firebase = await Firebase.initializeApp();

  return await (AppBootstrap(
    identifier: 'TEST',
    firebaseApp: firebase,
    appTitle: 'TEST',
    appInfo: TestAppInfoModule(),
    deviceInfo: TestDeviceInfoModule(),
    userPrefs: TestUserPrefsModule(),
    localization: TestLocalizationModule(),
    routing: FluroRoutingModule(
      historyObservers: historyObservers,
      initialRoute: initialPage ?? '/page1',
      routes: router,
      path: '/page1',
    ),
  ).buildRootWidget());
}

FRouter createMainRouter() {
  final FRouter routes = newTestRouter();

  routes.page("/page1", (context, parameters) => RouteTestPage.ofTabs(thisPage: 1, pageCount: 3));
  routes.page("/page2", (context, parameters) => RouteTestPage.ofTabs(thisPage: 2, pageCount: 3));
  routes.page("/page3", (context, parameters) => RouteTestPage.ofTabs(thisPage: 3, pageCount: 3));

  return routes;
}
