import 'dart:core';
import 'dart:io';

import 'package:http/io_client.dart';

import 'generate_sunny_icon_snippet.dart';
import 'icon_shared.dart';

/// Downloads all the sunny icons defined in the google doc
///
/// To generate all icons, use the fastlane command [fastlane icons]
///
/// Also prints out a snippet of code that for now is copied manually into sunny_icons.dart
main() async {
  final outputDir = Directory("./iconsource/svg");
  print("With output directory: $outputDir");
  if (outputDir.existsSync()) {
    print(" - Clearing existing output directory");
    outputDir.deleteSync(recursive: true);
  }
  print(" - Creating fresh directory");
  outputDir.createSync(recursive: true);

//  final icons =
//      Line8IconsSource("kevin@sunnyapp.co", "Pa\$\$word", outputDir);

  final icons = LocalFolderSource(
      targetDirectory: outputDir,
      sourceDirectory: Directory("./iconsource/input"));
  final prep = await icons.prepare();
  print(generateSunnyIconSnippet(prep.definitions));
}

abstract class IconSource {
  Future<IconPreparation> prepare();
}

class IconPreparation {
  final Directory? output;
  final List<IconDefinition> definitions;

  const IconPreparation(this.output, this.definitions);
}

class LocalFolderSource implements IconSource {
  final Directory? sourceDirectory;
  final Directory? targetDirectory;

  LocalFolderSource({this.sourceDirectory, this.targetDirectory});
  @override
  Future<IconPreparation> prepare() async {
    final items = await sourceDirectory!
        .list(recursive: true)
        .where((entity) => entity is File && entity.path.endsWith(".svg"))
        .map((entity) {
      final file = entity as File;

      final name = entity.absolute.path.split("/").last;
      final readableName = name.replaceAll(".svg", "");
      final fName = readableName.replaceAll(" ", "_").toLowerCase();
      file.copy("${targetDirectory!.path}/$fName.svg");
      return IconDefinition(
          iconName: readableName,
          iconVariableName: fName,
          iconOutlineUri: Uri.parse("file:///${file.path}"),
          iconSolidUri: Uri.parse("file:///${file.path}"));
    }).toList();

    return IconPreparation(sourceDirectory, items);
  }
}

class Line8IconsSource implements IconSource {
  final String username;
  final String password;
  final Directory outputDir;
  final client = IOClient(HttpClient()..findProxy = null);

  Line8IconsSource(this.username, this.password, this.outputDir);

  Future<IconPreparation> prepare() async {
    print("Authenticating... against line8");
    final apiKey = await authenticate(username, password);
    print("Authenticated with key $apiKey");

    downloadIcon(line8Id, IconType type, iconName, {int? size}) async {
      final fontName =
          "${iconName}_${type == IconType.outline ? "outline" : "solid"}";
      final params = <String, String>{
        "id": "$line8Id",
        "color": "000000",
        "format": "svg",
        "token": "$apiKey",
      };
      if (size != null) {
        params["size"] = "$size";
      }
      File file = File("${outputDir.path}/$fontName.svg");
      final requestUri = Uri.https("img.icons8.com", "/", params);
      print("Download uri: $requestUri");
      await client.get(requestUri).then((resp) {
        if (resp.statusCode != 200) {
          throw "Invalid response for icon ${resp.statusCode}: ${resp.body}";
        }
        print("Writing $iconName to $file");
        final bytes = resp.bodyBytes;
        file.writeAsBytesSync(bytes, flush: true);
      });
      return fontName;
    }

    final iconDefinitions = await readIconDefinitions(client);
    for (final icon in iconDefinitions) {
      await Future.wait([
        downloadIcon(icon.outlineUriId, IconType.outline, icon.iconName),
        downloadIcon(icon.solidUriId, IconType.solid, icon.iconName),
      ]);
    }

    return IconPreparation(outputDir, [...iconDefinitions]);
  }
}
